#include <stdio.h>
#include <stdlib.h>
#include <gsl/gsl_matrix.h>

// Return a random number in [0;1[
double rnd(void)
{
    return (double)rand()/RAND_MAX;
}

// Initialize matrix M with random ints from low to high
void init_rnd_matrix(gsl_matrix* M, double low, double high)
{
    int row, col;
    for (row = 0; row < M->size1; row++)
        for (col = 0; col < M->size2; col++)
            gsl_matrix_set(M, row, col, 
                    (int)(rnd() * (high-low) + low));
}

// Write matrix to stdout
void print_matrix(gsl_matrix* M)
{
    int row, col;
    double val;
    for (row = 0; row < M->size1; row++) {
        for (col = 0; col < M->size2; col++) {
            val = gsl_matrix_get(M, row, col);
            printf("%f\t", val);
        }
        puts(""); // newline
    }
}

// Add a slab of sand from the target cell
void deposit(gsl_matrix* Z, int row, int col)
{
    gsl_matrix_set(Z, row, col, gsl_matrix_get(Z, row, col) + 1.0);
}

// Remove a slab of sand from the target cell
void erode(gsl_matrix* Z, int row, int col)
{
    gsl_matrix_set(Z, row, col, gsl_matrix_get(Z, row, col) - 1.0);
}

// Find out if cell is in shade (1) or not (0)
int inshade(
        gsl_matrix* Z,  // matrix with sand slabs
        int row,        // row idx of interest
        int col)        // col idx of interest
{
    int js;
    int i = 1;
    int shade = 0;
    int nx = Z->size2;

    while ((i <= nx/4) && (shade == 0)) {
        js = col - i;
        if (js < 0) // periodic boundary
            js += nx;
        if (gsl_matrix_get(Z, row, js) >= gsl_matrix_get(Z, row, col) + i)
            shade = 1;
        i++;
    }
    return shade;
}

void find_max_slope_neighbor_ero(
        gsl_matrix* Z,  // sand slab values
        int row,        // target row
        int col,        // target col
        int* row_max,   // max slope neighbor row
        int* col_max,   // max slope neighbor col
        double* dh_max) // max slope
{

    int n;  // 1d neighbor idx
    int i, j; // 2d neighbor idx
    double dh;  // slope
    
    // relative neighbor coordinates and distances
    int drow[] = {1, 1, 1, 0, -1, -1, -1, 0};
    int dcol[] = {-1, 0, 1, 1, 1, 0, -1, -1};
    double dx[] = {1.4142, 1.0, 1.4142, 1.0, 1.4142, 1.0, 1.4142, 1.0};

    // matrix dimensions
    int nrows = Z->size1;
    int ncols = Z->size2;

    // check the 8 surrounding neighbor cells
    for (n = 0; n<8; n++) {
        
        // absolute neighbor cell idx
        i = row + drow[n];
        j = col + dcol[n];

        // correct for periodic boundaries
        if (i < 0)
            i += nrows;
        if (i >= nrows)
            i -= nrows;
        if (j < 0)
            j += ncols;
        if (j >= ncols)
            j -= ncols;

        // find slope
        dh = (gsl_matrix_get(Z, i, j) - gsl_matrix_get(Z, row, col)) / dx[n];

        // save position if it is the highest slope so far
        if (dh > *dh_max) {
            *dh_max = dh;
            *row_max = i;
            *col_max = j;
        }
    }
}

void find_max_slope_neighbor_depo(
        gsl_matrix* Z,  // sand slab values
        int row,        // target row
        int col,        // target col
        int* row_max,   // max slope neighbor row
        int* col_max,   // max slope neighbor col
        double* dh_max) // max slope
{

    int n;  // 1d neighbor idx
    int i, j; // 2d neighbor idx
    double dh;  // slope
    
    // relative neighbor coordinates and distances
    int drow[] = {1, 1, 1, 0, -1, -1, -1, 0};
    int dcol[] = {-1, 0, 1, 1, 1, 0, -1, -1};
    double dx[] = {1.4142, 1.0, 1.4142, 1.0, 1.4142, 1.0, 1.4142, 1.0};

    // matrix dimensions
    int nrows = Z->size1;
    int ncols = Z->size2;

    // check the 8 surrounding neighbor cells
    for (n = 0; n<8; n++) {
        
        // absolute neighbor cell idx
        i = row + drow[n];
        j = col + dcol[n];

        // correct for periodic boundaries
        if (i < 0)
            i += nrows;
        if (i >= nrows)
            i -= nrows;
        if (j < 0)
            j += ncols;
        if (j >= ncols)
            j -= ncols;

        // find slope
        dh = (gsl_matrix_get(Z, row, col) - gsl_matrix_get(Z, i, j)) / dx[n];

        // save position if it is the highest slope so far
        if (dh > *dh_max) {
            *dh_max = dh;
            *row_max = i;
            *col_max = j;
        }
    }
}

// Check and perform avalanche into cell if slope exceeds limit
void avalanche_erosion(
        gsl_matrix* Z,  // sand slab values
        int row,        // target row
        int col)        // target col
{
    // find the neighbor with the max. height difference and the slope value
    double dh_max = 0.0;    // max slope
    int row_max, col_max;
    find_max_slope_neighbor_ero(Z, row, col, &row_max, &col_max, &dh_max);

    // Perform avalanche along max slope neighbor
    double threshold = 2.0; // avalanche threshold
    if (dh_max > threshold) {
        deposit(Z, row, col); // put sand in target cell
        erode(Z, row_max, col_max); // remove sand from max slope neighbor cell

        // Recursion
        avalanche_erosion(Z, row, col);
    }
}

// Check and perform avalanche away from cell if slope exceeds limit
void avalanche_deposition(
        gsl_matrix* Z,  // sand slab values
        int row,        // target row
        int col)        // target col
{
    // find the neighbor with the max. height difference and the slope value
    double dh_max = 0.0;    // max slope
    int row_max, col_max;
    find_max_slope_neighbor_depo(Z, row, col, &row_max, &col_max, &dh_max);

    // Perform avalanche along max slope neighbor
    double threshold = 2.0; // avalanche threshold
    if (dh_max > threshold) {
        erode(Z, row, col);
        deposit(Z, row_max, col_max);

        // Recursion
        avalanche_deposition(Z, row, col);
    }
}

// Move a sand unit and deposit it where it fits
void move_and_deposit(
        gsl_matrix* Z,  // matrix with sand slabs
        double p_ns,    // likelihood of deposition in no-sand cell
        double p_s,     // likelihood of deposition in sand cell
        double d_l,     // transport distance in no. of cells
        int* row,       // current cell row
        int* col,       // current cell col
        int* deposited) // deposited? 1=yes, 0=no
{
    // move sand slab in wind direction
    *col += d_l;
    int ncols = Z->size1;
    if (*col >= ncols)
        *col -= ncols;

    // is the target in the shade?
    // 1=yes, 0=no
    int shade = 0;
    shade = inshade(Z, *row, *col);
    if (shade > 0) {
        deposit(Z, *row, *col);
        avalanche_deposition(Z, *row, *col); // check for avalanches
        *deposited = 1; // sand deposited, stop moving it
    }

    // if not in the shade, check against deposition probabilities
    else {

        // sand in cell
        if (gsl_matrix_get(Z, *row, *col) > 0.) {
            if (rnd() <= p_s) { // deposit in cell with sand
                deposit(Z, *row, *col);
                avalanche_deposition(Z, *row, *col); // check for avalanches
                *deposited = 1; // sand deposited, stop moving it
            }
        } else { // no sand in cell
            if (rnd() <= p_ns) { // deposit in cell without sand
                deposit(Z, *row, *col);
                avalanche_deposition(Z, *row, *col); // check for avalanches
                *deposited = 1; // sand deposited, stop moving it
            }
        }
    }
}

// Perform a single Werner iteration
void werner_iter(
        gsl_matrix* Z,  // matrix with sand slabs
        int d_l,        // wind transport distance
        double p_ns,    // likelihood of deposition in sand-free cell
        double p_s)     // likelihood of deposition in sand-covered cell
{
    // Evaluate N=nx*ny cells in random order
    int row, col, deposited;
    int nrows = Z->size1; // row major
    int ncols = Z->size2;
    long int n;
    long int N = nrows*ncols;
    double cellval;


    // Perform cycle
    for (n=0; n<N; n++) {

        // random cell idx
        row = rand()%nrows;
        col = rand()%ncols;

        // If the cell has sand in it (val>0), and is not in the shadow zone
        cellval = gsl_matrix_get(Z, row, col);
        if ((cellval > 0.) && (inshade(Z, row, col) == 0)) {

            // erode
            erode(Z, row, col);

            // check for avalanche
            avalanche_erosion(Z, row, col);

            // move eroded sand and deposit where it fits
            deposited = 0;
            while (deposited == 0)
                move_and_deposit(Z, p_ns, p_s, d_l, &row, &col, &deposited);
                
        }
    }
}

// Loop system through time
void werner_loop(
        gsl_matrix* Z,  // matrix with sand slabs
        long int N_c,   // number of iterations
        int d_l,        // wind transport distance
        double p_ns,    // likelihood of deposition in sand-free cell
        double p_s)     // likelihood of deposition in sand-covered cell
{
    long int t;
    for (t = 0; t<N_c; t++)
        werner_iter(Z, d_l, p_ns, p_s);
}
